#include <errno.h>
#include <stdio.h>
#include <windows.h>

#ifdef __cplusplus
extern "C" {
#endif

static FILE *file;
static const unsigned long max_size = 35000;

/* Redirect stdout */
int WINAPI DllMain(HINSTANCE instance, unsigned long reason, void *crap) {
#ifndef USE_STDOUT
  file = fopen("flashpointserver.log", "ab");
  if (! file) {
    fprintf(file, "Can't open flashpointserver.log for writing: %s\n", strerror(errno));
    return 0;
  }
#else
  file = stdout;
#endif
  return 1;
}

/* Write current date and time to a buffer */
static int get_time(char *buffer, int buflen) {
  int size;
  char *date, *time;
  SYSTEMTIME now;

  /* Get current date and time */
  GetSystemTime(&now);

  /* Format date */
  size = GetDateFormat(LOCALE_SYSTEM_DEFAULT, 0, &now, "yyyy-MM-dd", 0, 0);
  date = (char *) malloc(size);
  if (! GetDateFormat(LOCALE_SYSTEM_DEFAULT, 0, &now, "yyyy-MM-dd", date, size)) {
    fprintf(file, "Error getting current date: %d\n", GetLastError());
    free(date);
    return 1;
  }

  /* Format time */
  size = GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, &now, "HH:mm:ss", 0, 0);
  time = (char *) malloc(size);
  if (! GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, &now, "HH:mm:ss", time, size)) {
    fprintf(file, "Error getting current time: %d\n", GetLastError());
    free(date);
    free(time);
    return 2;
  }

  /* Print */
  if (_snprintf(buffer, buflen, "%s %s", date, time) < 0) {
    fprintf(file, "Error writing date and time to buffer: %d\n", GetLastError());
    free(date);
    free(time);
    return 3;
  }

  /* Cleanup */
  free(date);
  free(time);

  return 0;
}

/* Replacement logging function */
__declspec(dllexport) void __stdcall furrycat_log(char *message) {
  char timestamp[32];

  /* Get timestamp for logging */
  if (get_time(timestamp, sizeof(timestamp))) timestamp[0] = '\0';

  /* Log the message */
  fprintf(file, "%s %s\n", timestamp, message);
  fflush(file);
}

/* Check the requested filename is a subdirectory of tmp2303/... */
static int is_sounds_dir(const char *s) {
  /* start and end (will/may be) for logging player names */
  unsigned long i;

  i = 0;

  if (s[i] != 't' && s[i] != 'T') return 0; i++;
  if (s[i] != 'm' && s[i] != 'M') return 0; i++;
  if (s[i] != 'p' && s[i] != 'P') return 0; i++;
  if (s[i] != '2') return 0; i++;
  if (s[i] != '3') return 0; i++;
  if (s[i] != '0') return 0; i++;
  if (s[i] != '2') return 0; i++;
  if (s[i] != '\\') return 0; i++;
  if (s[i] != 'p' && s[i] != 'P') return 0; i++;
  if (s[i] != 'l' && s[i] != 'L') return 0; i++;
  if (s[i] != 'a' && s[i] != 'A') return 0; i++;
  if (s[i] != 'y' && s[i] != 'Y') return 0; i++;
  if (s[i] != 'e' && s[i] != 'E') return 0; i++;
  if (s[i] != 'r' && s[i] != 'R') return 0; i++;
  if (s[i] != 's' && s[i] != 'S') return 0; i++;
  if (s[i] != '\\') return 0; i++;
  /* Skip player name */
  for ( ; s[i] != '\\'; i++);
  if (s[i] != '\\') return 0; i++;
  if (s[i] != 's' && s[i] != 'S') return 0; i++;
  if (s[i] != 'o' && s[i] != 'O') return 0; i++;
  if (s[i] != 'u' && s[i] != 'U') return 0; i++;
  if (s[i] != 'n' && s[i] != 'N') return 0; i++;
  if (s[i] != 'd' && s[i] != 'D') return 0; i++;
  if (s[i] != '\\') return 0; i++;

  /* Now i is offset of last backslash in 'tmp2302\players\xxx\sound\...' */
  return i;
}

/* Check the size of files in a sound directory - return 0 if full */
static int under_limit(char *dir, unsigned long max_size) {
  WIN32_FIND_DATA found;
  HANDLE handle;
  unsigned long size = 0;
  char path[MAX_PATH];

  /* Look for files in the directory */
  if (_snprintf(path, sizeof(path), "%s/*", dir) < 0) {
    fprintf(file, "snprintf() error: %d\n", GetLastError());
    /* Error so signal that it's OK to write files in here */
    return -1;
  }

  found.dwFileAttributes = FILE_ATTRIBUTE_NORMAL;
  handle = FindFirstFile(path, &found);
  if (handle == INVALID_HANDLE_VALUE) return -2;

  do {
    /* Check size */
    if (found.nFileSizeHigh) {
      /* Force overrun if file is that big! */
      size += max_size;
      break;
    }

    /* Add size */
    size += found.nFileSizeLow;
  } while (FindNextFile(handle, &found));

  /* Clean up */
  FindClose(handle);

  /* Lock the directory if it's too big */
  if (size > max_size) return 0;

  return 1;
}

/* Override CreateFileA */
__declspec(dllexport) HANDLE WINAPI furrycat_CreateFileA(IN LPCSTR lpFileName, IN DWORD dwDesiredAccess, IN DWORD dwShareMode, IN LPSECURITY_ATTRIBUTES lpSecurityAttributes, IN DWORD dwCreationDisposition, IN DWORD dwFlagsAndAttributes, IN HANDLE hTemplateFile) {
  unsigned long i;
  char *path, *player;
  char blurb[4096];

  /* Only interested in WRITING files */
  if (! (dwDesiredAccess & GENERIC_WRITE)) goto passthrough;

  /* Only interested in writing to sounds directory */
  i = is_sounds_dir(lpFileName);
  if (! i) goto passthrough;

  /* Allocate memory for copy of path name */
  path = (char *) HeapAlloc(GetProcessHeap(), 0, i);
  if (! path) goto passthrough;

  /* Copy path */
  CopyMemory(path, lpFileName, i);
  path[i - 1] = '\0';

  /* Check the directory */
  if (under_limit(path, max_size)) goto cleanup_passthrough;

  /* Extract player name from path */
  player = path + i;
  for ( ; *player != '\\'; player--); *player = '\0';
  for ( ; *player != '\\'; player--); player++;

  /* Log */
  _snprintf(blurb, sizeof(blurb), "Refusing sound \"%s\" for player %s", lpFileName + i, player);
  furrycat_log(blurb);

  /* Cleanup */
  HeapFree(GetProcessHeap(), 0, path);

  /* Deny */
  return INVALID_HANDLE_VALUE;

cleanup_passthrough:
  /* Cleanup */
  HeapFree(GetProcessHeap(), 0, path);

passthrough:
  return CreateFileA(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile);
}

#ifndef USR_DLL
int main(int argc, char **argv) {
  furrycat_CreateFileA(argv[1], GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
}
#endif

#ifdef __cplusplus
}
#endif
